local tinsert, pairs, assert = tinsert, pairs, assert;
local CreateFrame = CreateFrame;
local after = C_Timer.After;
local LE_QUEST_FREQUENCY_DAILY, LE_QUEST_FREQUENCY_WEEKLY = LE_QUEST_FREQUENCY_DAILY, LE_QUEST_FREQUENCY_WEEKLY;

local function merge(tableInWhichValuesWillBeInserted, tableContainingValuesToInsert)
	for key, value in pairs(tableContainingValuesToInsert)do
		tableInWhichValuesWillBeInserted[key] = value;
	end
end

Questo_API.dialogs.buttons = {};
local API = Questo_API.dialogs.buttons;
local Dialogs = Questo_API.dialogs;

local buttonsBag = {};

local DIALOG_CHOICE_BUTTON_MARGIN = 5;
local DIALOG_CHOICE_BUTTON_PADDING = 25;

local BUTTON_TEMPLATE = "Questo_DialogChoiceButton";
local defaultButton = CreateFrame("Button", nil, nil, BUTTON_TEMPLATE);

function API.getMargin()
	return DIALOG_CHOICE_BUTTON_MARGIN;
end

local GOSSIP_ICONS_TEXTURE_PATH = [[Interface\GossipFrame\%sGossipIcon]];

local function getIconTextureForGossipType(gossipType)
	return GOSSIP_ICONS_TEXTURE_PATH:format(gossipType);
end

local AVAILABLE_QUEST_ICONS_TEXTURE_PATHS = {
	DEFAULT   = [[Interface\GossipFrame\AvailableQuestIcon]],
	LEGENDARY = [[Interface\GossipFrame\AvailableLegendaryQuestIcon]],
	DAILY     = [[Interface\GossipFrame\DailyQuestIcon]]
}

local function getIconTextureForAvailableQuestType(frequency, isRepeatable, isLegendary)
	local questIcon = AVAILABLE_QUEST_ICONS_TEXTURE_PATHS.DEFAULT;
	if isLegendary then
		questIcon = AVAILABLE_QUEST_ICONS_TEXTURE_PATHS.LEGENDARY;
	elseif frequency == LE_QUEST_FREQUENCY_DAILY or frequency == LE_QUEST_FREQUENCY_WEEKLY or isRepeatable then
		questIcon = AVAILABLE_QUEST_ICONS_TEXTURE_PATHS.DAILY;
	end
	return questIcon;
end

local ACTIVE_QUEST_ICONS_TEXTURE_PATHS = {
	DEFAULT   = [[Interface\GossipFrame\ActiveQuestIcon]],
	LEGENDARY = [[Interface\GossipFrame\ActiveLegendaryQuestIcon]],
	DAILY     = [[Interface\GossipFrame\DailyActiveQuestIcon]]
}

local function getIconTextureForActiveQuestType(frequency, isRepeatable, isLegendary)
	local questIcon = ACTIVE_QUEST_ICONS_TEXTURE_PATHS.DEFAULT;
	if isLegendary then
		questIcon = ACTIVE_QUEST_ICONS_TEXTURE_PATHS.LEGENDARY;
	elseif frequency == LE_QUEST_FREQUENCY_DAILY or frequency == LE_QUEST_FREQUENCY_WEEKLY or isRepeatable then
		questIcon = ACTIVE_QUEST_ICONS_TEXTURE_PATHS.DAILY;
	end
	return questIcon;
end

local BUTTON_DECORATORS = {
	[Dialogs.BUCKET_TYPE.COMPLETED_QUEST] = function(button, data)
		button:SetText(data.title);
		button:SetIcon(getIconTextureForActiveQuestType(data.frequency, data.isRepeatable, data.isLegendary));
	end,
	[Dialogs.BUCKET_TYPE.AVAILABLE_QUEST] = function(button, data)
		button:SetText(data.title);
		button:SetIcon(getIconTextureForAvailableQuestType(data.frequency, data.isRepeatable, data.isLegendary));
		if data.isTrivial then
			button:GreyOutIcon();
		end
	end,
	[Dialogs.BUCKET_TYPE.GOSSIP] = function(button, data)
		button:SetText(data.title);
		button:SetIcon(getIconTextureForGossipType(data.gossipType));
	end,
	[Dialogs.BUCKET_TYPE.UNCOMPLETED_QUEST] = function(button, data)
		button:SetText(data.title);
		button:SetIcon("Interface\\GossipFrame\\IncompleteQuestIcon");
	end
}

local BUTTON_API = {
	RefreshHeight = function(button)
		button:SetHeight(button.text:GetHeight() + DIALOG_CHOICE_BUTTON_PADDING);
	end,
	SetText = function(button, text)
		button.text:SetText(text);
		button:RefreshHeight();
	end,
	FadeIn = function(button)
		button:Show();
		button.fadeIn:Play();
	end,
	SetIcon = function(button, texture)
		button.icon:SetTexture(texture);
	end,
	GreyOutIcon = function(button)
		button.icon:SetVertexColor(0.5, 0.5, 0.5);
	end,
	UnGreyOutIcon = function(button)
		button.icon:SetVertexColor(1, 1, 1);
	end,
	SetBindingText = function(button, bindingText)
		button.binding:SetText(bindingText);
		button.binding:Show();
	end,
	HideBidingText = function(button)
		button.binding:SetText("");
		button.binding:Hide();
	end,
	PlayPlayerModelAnimation = function(button)
		local animation;
		local playerModel = Questo_NPCFrame.models.me;
		button.text:GetText():gsub("[%.%?%!]+", function(finder)
		end);
	end,
	Decorate = function(self, buttonIndex, choiceIndex, data, bucketType, eventType)

		assert(BUTTON_DECORATORS[bucketType], ("No button decorator available for bucket type %s!"):format(bucketType or "NO_BUCKET_TYPE"));

		if buttonIndex < 10 and Questo_Data.config.useKeyboard then
			self:SetBindingText(buttonIndex);
			self.bindingKey = buttonIndex;
		else
			self:HideBidingText();
		end


		BUTTON_DECORATORS[bucketType](self, data);
		self:RefreshHeight();


		self:SetScript("OnClick", function()
			Dialogs.getDialogChoiceSelectorForEventType(eventType, bucketType)(choiceIndex);
		end)

		after(0.05 * buttonIndex, function()
			self:FadeIn();
		end);
	end,
	SetFont = function(self, ...)

	end,
	Release = function(self)
		self.isAvailable = true;
		self:UnGreyOutIcon();
		self:Hide();
	end
}

local DEFAULT_ANCHOR_POINT = "BOTTOM";
local FIRST_OPTION_ANCHOR_POINT = "TOP";

function API.getButton(parent, anchor)

	local button;
	local anchorPoint = anchor and DEFAULT_ANCHOR_POINT or FIRST_OPTION_ANCHOR_POINT;
	if not anchor then
		anchor = parent;
	end

	for _, existingButton in pairs(buttonsBag) do
		if existingButton.isAvailable then
			existingButton.isAvailable = false;
			button = existingButton;
			break;
		end
	end

	if not button then
		button = CreateFrame("Button", nil, parent, BUTTON_TEMPLATE);
		merge(button, BUTTON_API);

		local fontSettings = API.getFontSettings();
		button:SetFont(fontSettings.font, fontSettings.size, fontSettings.outline);

		tinsert(buttonsBag, button);
	end

	button:SetPoint("TOP", anchor, anchorPoint, 0, -1 * DIALOG_CHOICE_BUTTON_MARGIN);

	return button;
end

function API.hideAllButtons()
	for _, existingButton in pairs(buttonsBag) do
		existingButton:Release();
	end
end

function API.selectOptionAtIndex(buttonIndex)
	local foundOptionForIndex = false;
	for _, existingButton in pairs(buttonsBag) do
		if existingButton.bindingKey == buttonIndex and existingButton:IsShown() then
			existingButton:Click();
			foundOptionForIndex = true;
		end
	end
	return foundOptionForIndex;
end

local DEFAULT_FONT_SETTINGS = { defaultButton.text:GetFont() };
local fontSettings = {};
defaultButton:Hide();

local function applyFontStyleToEveryExistingButtons()
	local fontSettings = API.getFontSettings();
	local totalButtonHeights = 0;
	for _, existingButton in pairs(buttonsBag) do
		existingButton:SetFont(fontSettings.font, fontSettings.size, fontSettings.outline);
		existingButton:RefreshHeight();
		totalButtonHeights = totalButtonHeights + existingButton:GetHeight() + API.getMargin();
	end
	Questo_API.dialogs.scrollFrame.show(totalButtonHeights);
end

function API.getDefaultFontStyle()
	return DEFAULT_FONT_SETTINGS;
end

function API.getFontSettings()
	return fontSettings;
end

function API.setFontStyle(fontStyle)
	fontSettings = {
		font = Questo_API.lib.getFontPath(fontStyle.Font),
		size = fontStyle.Size,
		outline = fontStyle.Outline
	}
	applyFontStyleToEveryExistingButtons();

end